/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - solaris_kstat.c */
/* rdk@krupczak.org */

/* Solaris kstat and other kernel-related functions */
/* can be linked with other plugins */


#ifdef SOLARIS

#include <kstat.h>  
#include <stdio.h>
#include <inttypes.h>
#include <stdlib.h>
#include <sys/types.h>
#include <strings.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "xmp.h"

static kstat_ctl_t *kc = NULL;

/* ************************************************************* */
/* given module, instance, name, fetch kstat and return 
   value in provided space of length=buflen.  Caller really
   has to know a prior the type/size of the returned value 
   in order for this call to be effectively used.  Size
   of value is returned.
*/

int get_named_kstat(char *mod, int inst, char *name, char *stat, 
                    void *buf,int buflen)
{
     kstat_t *ksp;
     kstat_named_t *knp;
     int32_t aInt;
     uint32_t aUint;
     int64_t a64int;
     uint64_t aU64int;

     if ((buf == NULL) || (buflen < 4)) {
         /* fprintf(stderr,"get_named_kstat: bad buf or buflen\n"); */
        return -1;  
     }
     if (stat == NULL) {
        /* fprintf(stderr,"get_named_kstat: no statname\n"); */
        return -1;  
     }

     if (kc == NULL) {
        if ((kc = kstat_open()) == NULL) {
	   /* fprintf(stderr,"get_named_kstat: bad buf or buflen\n"); */
           return -1;
        }
     }
     kstat_chain_update(kc);
     if ((ksp = kstat_lookup(kc,NULL,0,name)) == NULL) {
         /* fprintf(stderr,"get_named_kstat: lookup failed\n"); */
         return -1;
     }
     if (kstat_read(kc,ksp,NULL) < 0) {
         /* fprintf(stderr,"get_named_kstat: read failed\n"); */
         return -1;
     }
     if ((knp = kstat_data_lookup(ksp,stat)) == NULL) {
         /* fprintf(stderr,"get_named_kstat: data lookup failed\n"); */
         return -1;
     }

     switch(knp->data_type) {  

       case KSTAT_DATA_INT32:
	    /* printf("32-%d\n",knp->value.i32); */
	    aInt = knp->value.i32;
            switch (buflen) {
 	      case 4:
	 	 (*(int32_t *)(buf)) = aInt;
                 return 4;
                 break;

  	      case 8:
	 	 (*(int64_t *)(buf)) = (int64_t)aInt;
                 return 8;
                 break;

	      default:
	 	 return -1;
            }
            break;  

       case KSTAT_DATA_UINT32:  
	    aUint = knp->value.ui32;
            /* printf("32-%d\n",knp->value.ui32); */
            switch (buflen) {
 	      case 4:
	 	 (*(uint32_t *)(buf)) = aUint;
                 return 4;
                 break;

  	      case 8:
	 	 (*(uint64_t *)(buf)) = (uint64_t)aUint;
                 return 8;
                 break;

	      default:
	 	 return -1;
            }
            break;  

       case KSTAT_DATA_INT64:  
   	    a64int = knp->value.i64;
            /* printf("64-%lld\n",knp->value.i64); */
            switch (buflen) {

 	      case 4: /* may truncate ! */
	 	 (*(int32_t *)(buf)) = (int32_t)a64int;
                 return 4;
                 break;

  	      case 8:
	 	 (*(int64_t *)(buf)) = a64int;
                 return 8;
                 break;

	      default:
		 return -1;
            }
            break;  

       case KSTAT_DATA_UINT64:  
   	    aU64int = knp->value.ui64;
            /* printf("64-%lld\n",knp->value.ui64); */
            switch (buflen) {
 	      case 4:
	 	 (*(uint32_t *)(buf)) = (uint32_t)aU64int;
                 return 4;
                 break;

  	      case 8:
	 	 (*(uint64_t *)(buf)) = aU64int;
                 return 8;
                 break;

	      default:
		 return -1;
            }
            break;

       case KSTAT_DATA_CHAR:
          /* char values */
  	  aInt = 16;
	  if (buflen < 16)
	     aInt = buflen;
          bcopy(knp->value.c,buf,aInt);
          ((char *)buf)[aInt-1] = 0;
          return strlen(buf);
	  break;

       case KSTAT_DATA_STRING:
	  aInt = knp->value.string.len;  
          if (buflen < aInt)
 	     aInt = buflen;
          strncpy(knp->value.string.addr.ptr,buf,aInt);
          ((char *)buf)[aInt-1] = 0;
          return aInt;
  	  break;

       default:
 	    /* we dont support these types */
	    return -1;
     }  
}

/* ************************************************************* */
/* use this function to obtain raw kstat blocks containing well-known
   structures (e.g. vminfo); caller has to know the
   name and instance and provide space to copy into 
 */

int get_raw_kstat(char *module, char *name, int inst, void *buf, int maxlen)
{
     kstat_t *ksp;
     int retlen;

     if ((name == NULL) || (buf == NULL))
        return 0;

     if (kc == NULL) {
        if ((kc = kstat_open()) == NULL) {
	   /* fprintf(stderr,"get_named_kstat: bad buf or buflen\n"); */
           return -1;
        }
     }

     kstat_chain_update(kc);

     /* find the one we want */
     for (ksp = kc->kc_chain; ksp != NULL; ksp = ksp->ks_next) {

         if (ksp->ks_type == KSTAT_TYPE_RAW) {

	    if ((module != NULL) && (strcmp(ksp->ks_module,module) != 0))
	       continue;
            if (strcmp(ksp->ks_name,name) != 0)
	       continue;
            if (ksp->ks_instance != inst)
	       continue;

  	   /* we found it */
           kstat_read(kc,ksp,NULL);

	   retlen = ksp->ks_data_size;
           if (retlen > maxlen)
	      retlen = maxlen;
           bcopy(ksp->ks_data,buf,retlen);
           return retlen;

	 } /* if raw */

     }

     return 0;
}

/* ************************************************************* */
/* called to close things up if plugin is unloaded */

int kstat_finish()
{

     if (kc != NULL)
        kstat_close(kc);
     kc = NULL;
}

/* ************************************************************* */

#endif /* SOLARIS */
