/* 
   COPYRIGHT 2005,7 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - coreModules.c */
/* rdk@krupczak.org */

/* 
   modules/files/plugins/dlls/libs that the core agent is tracking
   for purposes of upgrading, distributing, etc.
*/

#ifdef WIN32
#include <windows.h>
#include <Iphlpapi.h>
#include <io.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>

#ifndef WIN32
#include <arpa/inet.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <dlfcn.h>
#include <dirent.h>
#include <signal.h>
#include <sys/param.h>
#endif

#include "xmp.h"
#include "socklib.h"
#include "globals.h"
#include "xmp_vars.h"
#include "plugins.h"
#include "linkedlist.h"
#include "xmp_session.h"
#include "modules.h"
#include "pluginexport.h"

static linkedList_t *moduleList = NULL;
static int init = 0;

/* do we need to restart ourselves and perform an upgrade? */
static int upgradeNeeded = 0;
static time_t upgradeNeededTime = 0;

/* ************************************************************* */
/* write bufsize bytes from buf to filename */

#ifndef WIN32

int write_file(char *filename, u_char *buf, int bufsize)
{
     int fd;

     if ((filename == NULL) || (buf == NULL) || (bufsize < 1))
        return -1;

     if ((fd = open(filename,O_CREAT|O_WRONLY|O_TRUNC,0644)) < 0) {
        perror("open/create failed");
        return -1;
     }

     if (write(fd,(void *)buf,bufsize) != bufsize) {
        perror("write buf failed");
        close(fd);
        return -1;
     }

     close(fd);

     /* depending on file type, we should set settings appropriately 
        default to be 0644
        .so's should be 0755, .sh should be 0755, exe 0755, 
        init scripts should be 0755
      */

     /* permissions should probably be incorporated into modules
        table and passed as an attribute XXX */

     chmod(filename,0644);

     /* by default, make all non-xml files be 0755 */
     if (strstr(filename,".xml") == NULL)
        chmod(filename,0755);

     return bufsize;
}

#else

int write_file(char *filename, u_char *buf, int bufsize)
{
     int fd;

     if ((filename == NULL) || (buf == NULL) || (bufsize < 1))
        return -1;

     if ((fd = _open(filename,_O_CREAT|_O_WRONLY|_O_TRUNC|_O_BINARY,0644))<0) {
        fprintf(stderr,"write_fileopen/create failed\n");
        return -1;
     }

     if (_write(fd,(void *)buf,bufsize) != bufsize) {
        fprintf(stderr,"write_file: write buf failed\n");
        _close(fd);
        return -1;
     }

     _close(fd);

     /* should we set permissions on files written on win32? XXX */

     return bufsize;
}

#endif /* !WIN32 */

/* ************************************************************* */
/* read a file and return its contents into a single buffer */
/* no zero termination! */

#ifndef WIN32

u_char *read_file(char *filename, int *filelength)
{
     struct stat statbuf;
     u_char *returnbuf;
     int fd,ret;

     if ((filename == NULL) || (filelength == NULL))
        return NULL;

     if (stat(filename,&statbuf) < 0) {
        return NULL;
     }

     *filelength = (int)statbuf.st_size;

     if ((returnbuf = (u_char *)malloc((*filelength)+1)) == NULL) {
        return NULL;
     }

     if ((fd = open(filename,O_RDONLY)) < 0) {
        perror("open failed");
        free(returnbuf);
        return NULL;
     }

     if ((ret = read(fd,(void *)returnbuf,*filelength)) != *filelength) {
        fprintf(stderr,"read did not match %d,%d\n",ret,*filelength);
        *filelength = 0;
        free(returnbuf);
        close(fd);
        return NULL;
     }

     close(fd);
     return returnbuf;
}

time_t get_file_time(char *filename)
{
     struct stat statbuf;

     if (filename == NULL)
        return 0;

     if (stat(filename,&statbuf) < 0) {
        return 0;
     }
     if (statbuf.st_mtime > statbuf.st_ctime)
        return statbuf.st_mtime;

     return statbuf.st_ctime;
}

#else

u_char *read_file(char *filename, int *filelength)
{
     struct _stat statbuf;
     u_char *returnbuf;
     int fd,ret;

     if ((filename == NULL) || (filelength == NULL))
        return NULL;

     if (_stat(filename,&statbuf) < 0) {
        return NULL;
     }

     *filelength = (int)statbuf.st_size;

     if ((returnbuf = (u_char *)malloc((*filelength)+1)) == NULL) {
        return NULL;
     }

     if ((fd = _open(filename,_O_RDONLY|_O_BINARY)) < 0) {
        fprintf(stderr,"read_file: open failed\n");
        free(returnbuf);
        return NULL;
     }

     if ((ret = _read(fd,(void *)returnbuf,*filelength)) != *filelength) {
        fprintf(stderr,"read did not match %d,%d\n",*filelength,ret);
        *filelength = 0;
        free(returnbuf);
        _close(fd);
        return NULL;
     }

     _close(fd);
     return returnbuf;
}

time_t get_file_time(char *filename)
{
     struct _stat statbuf;

     if (filename == NULL)
        return 0;

     if (_stat(filename,&statbuf) < 0) {
        return 0;
     }
     if (statbuf.st_mtime > statbuf.st_ctime)
        return statbuf.st_mtime;

     return statbuf.st_ctime;
}

#endif /* !WIN32 */

/* ************************************************************* */
/* 
   get module contents, base64 encode them, and return
   them in buffer that we allocate; caller is responsible
   for free'ing this buffer, perhaps via xmpPduCleanup
   the returned buffer must be zero terminated but that
   zero termination is not part of length; that way,
   buildXmp
*/

u_char *getModuleContents(char *filename, int *contentSize)
{
     u_char *fileBuf;
     int fileBufSize;
     u_char *encodedBuf;
     int encodedBufSize;
     char *base64_encode(u_char *, int, int *);

     if ((filename == NULL) || (contentSize == NULL))
        return NULL;

     /* default */
     *contentSize = 0;

     /* read the file first into a temp buffer */
     fileBuf = read_file(filename,&fileBufSize);
     if (fileBuf == NULL) {
        return NULL;
     }

     /* encode into new buffer */
     encodedBufSize = 0;
     encodedBuf = (u_char *)base64_encode(fileBuf,fileBufSize,&encodedBufSize);

     /* free first buffer */
     if (fileBuf != NULL)
        free(fileBuf);

     /* set return content size */
     *contentSize = encodedBufSize;

     return encodedBuf;
}

/* ************************************************************* */
/* given base64 encoded module contents, decode, write to
   file, and return size of written file (decoded contents)
   we append '.new' to the module name
*/

int writeModuleContents(char *modulename, u_char *ebuf, int ebufsize, 
                        int modtype, char *newversion)
{
     char filename[256];
     int ret;
     u_char *decodeBuf;
     int decodeBufSize;
     u_char *base64_decode(u_char *, int, int *);

     if ((modulename == NULL) || (ebuf == NULL) || (ebufsize < 1))
        return -1;

     /* decode the base64 encoded buffer */
     decodeBufSize = 0;
     decodeBuf = base64_decode(ebuf,ebufsize,&decodeBufSize);
     if (decodeBuf == NULL) {
        return -1;
     }
    
     /* construct filename = modulename.new */
     sprintf(filename,"%s.new",modulename);

     /* write the file out */
     ret = write_file(filename,decodeBuf,decodeBufSize);

     /* write out the file type? NO
        depending on the file type, we need to do something 
        once we do this, we need to update our entry in the
        module table to reflect the newer version
     */

     /* done, cleanup */
     if (decodeBuf != NULL)
        free(decodeBuf);

     return decodeBufSize;
}

/* ************************************************************* */
/* query/retrieve the module contents (base64 format), decode
   and then write to file; return -1 or number bytes written */

int fetchModule(char *peerName, xmpSession_t *session, char *modulename, 
                int moduleType, char *newversion)
{
     xmpPdu_t query, reply;
     xmpVariable_t var1, *varptr;
     int ret;
     char linestr[1024];

     if ((peerName == NULL) || (session == NULL) || (modulename == NULL))
        return -1;

     if (xmpIsSessionConnected(session) == 0)
        return -1;

     fprintf(stderr,"fetchModule: going to get module '%s' from %s\n",
	     modulename,peerName);

     bzero((void *)&query,sizeof(query));
     query.msgType = SelectTableRequest;
     query.messageID = (int64_t)time(NULL);
     query.errorStatus = noError;
     query.errorIndex = 0;
     
     sprintf(query.tableMib,"core");
     sprintf(query.tableName,"moduleTable");
     sprintf(query.tableKey,"%s",modulename);

     bzero((void *)&var1,sizeof(var1));
     var1.next = NULL;
     sprintf(var1.varName.mibName,"core");
     sprintf(var1.varName.objName,"moduleContent");
     var1.varSyntax = nullSyntax;
     var1.valueStrLen = 0;

     query.numberVariables = 1;
     query.varList = &var1;
     query.varListTail = NULL;

     bzero((void *)&reply,sizeof(reply));

     if ((ret = xmpSendQueryReceiveReply(session,&query,&reply)) < 0) {
        fprintf(stderr,"fetchModule: module content query failed\n");
        xmpPduCleanup(&reply);
        return -1;
     }

     if ((reply.errorStatus == noError) && (reply.numberVariables == 1)) {

        varptr = reply.varList;

        if (varptr == NULL) {
           xmpPduCleanup(&reply);
           return -1;
        }
 
        DBG3("fetchModule: fetched %d bytes for %s from peer %s\n",
	      varptr->valueStrLen,modulename,peerName);
        sprintf(linestr,"fetchModule: fetched %d bytes for %s from peer %s\n",
		varptr->valueStrLen,modulename,peerName);
        /* xmpd_log(linestr); */

        /* decode and write the contents */
        ret = writeModuleContents(modulename,varptr->valueBuf,
                                  varptr->valueStrLen,moduleType,
                                  newversion);

     }

     xmpPduCleanup(&reply);

     return ret;
}

/* ************************************************************* */
/* add a module to the list that the agent tracks; if it already exists,
   update the fields were applicable
 */

int addModule(char *modulename, char *version, char *owner, 
              char *dependList, moduleType_t t)
{
     module_t comp, *m;
     linkedElement_t *e;

     /* if we're not initialized yet, dont worry */
     if (init == 0)
        return 0;

     if ((modulename == NULL) || (version == NULL) || (owner == NULL))
        return -1;

     if ((t < agentModule) || (t > binModule))
        return -1;

     /* if modulename already added, dont add again but update */
     strcpy(comp.modulename,modulename);
     if ((e = findElement(moduleList,&comp)) != NULL) {

        m = (module_t *)e->data;
        if (m == NULL)
	   return -1;

        /* leave module name unchanged but update other fields */
        if (version != NULL)
           strcpy(m->version,version);
        if (owner != NULL)
           strcpy(m->owner,owner);
        if (t != invalidModuleType)
           m->modType = t;
        if (dependList != NULL)
           strcpy(m->dependList,dependList);
        return 0;
     }

     if ((m = (module_t *)malloc(sizeof(module_t))) == NULL) {
         fprintf(stderr,"addModule: malloc failed, %d\n",errno);
         return -1;
     }

     strcpy(m->modulename,modulename);
     strcpy(m->version,version);
     strcpy(m->owner,owner);
     m->pendingVersion[0] = 0;
     m->modType = t;
     m->dependList[0] = 0;
     if (dependList != NULL)
        strcpy(m->dependList,dependList);

     /* find out create/modify time */
     m->timeStamp = 0;
     m->timeStamp = get_file_time(m->modulename);
 
     insertHead(moduleList,(void *)m);

     return 1;
}

/* ************************************************************* */
/* given a module, update its pending version; called by 
   compareModuleLists() since it is working with a copy of
   the actual core module list and not the actual list; by updating
   the pendingModule list, we wont go download other versions 
   that may be newer than the current one but older than the one
   we may have downloaded for updating  
*/
  
int updateModulePendingVersion(char *moduleName, float pendingVersion)
{
     module_t comp, *m;
     linkedElement_t *e;

     if ((moduleName == NULL) || (pendingVersion <= 0.0)) {
        return -1;
     }

     strcpy(comp.modulename,moduleName);
     if ((e = findElement(moduleList,&comp)) == NULL) {
        return -1;
     }

     if ((m = (module_t *)e->data) == NULL) {
        return -1;
     }

     sprintf(m->pendingVersion,"%3.2f",pendingVersion);
 
     return 1;
}

/* ************************************************************* */
/* remove a module entry from list we are tracking for
   software updating and distribution */

int removeModule(char *modulename)
{
     module_t m;
     linkedElement_t *e;

     /* find element containing modulename */
     strcpy(m.modulename,modulename);
     e = findElement(moduleList,(void *)&m);

     /* destroy it */
     if (e != NULL)
        destroyElement(moduleList,e);    

     return 1;
}

/* ************************************************************* */

int printModule(module_t *m)
{
     if (m == NULL)
        return -1;

     fprintf(stderr,"Moudle:%s Version:%s Owner:%s\n",
             m->modulename,m->version,m->owner);

     return 1;
}

int compareModules(module_t *m1, module_t *m2)
{
     if ((m1 == NULL) || (m2 == NULL))
        return -1;

     return strcmp(m1->modulename,m2->modulename);
}

int freeModule(module_t *m)
{
     if (m != NULL)
       free(m);
     return 1;
}

/* ************************************************************* */

xmpErrorStatus_t var_modList(char *key, xmpVariable_t *columns,
                                xmpPdu_t *out)
{
     char rowKey[256];
     xmpVariable_t *varptr, *cptr;
     u_int maxrows; /* important to be unsigned */
     int numRows = 0;
     int numVariables = 0;
     linkedElement_t *e;
     module_t *m;
     char timestr[32];

     if (xmpd_verbose)
        printf("var_modList: starting\n");

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     if (autoUpgrade != 1) {
        return noSuchName;
     }

     /* key is: moduleName -- keys can be name, partial, *, \0 */
     
     /* get pdu ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     /* lock modList linked list */
     pthread_mutex_lock(&moduleList->lock);

     /* cycle through entries */
     e = moduleList->head;
     while (e != NULL) {

       m = (module_t *)e->data;
       if (m == NULL) break;

       /* row key is module name */
       sprintf(rowKey,"%s",m->modulename);

       if ((strcmp(rowKey,key) != 0) && (key[0] != 0) && 
           (key[0] != '*')) {
	  e = e->next;
 	  continue; /* with while modules */
       }

       cptr = columns;
       while ((cptr != NULL) && (numRows < maxrows)) {
       
         /* malloc a return value */
         if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
             == NULL) {
	     perror("var_modList: malloc failed");
             out->errorStatus = genError;
             /* release lock */
             pthread_mutex_unlock(&moduleList->lock);
             return genError;
         }
         bcopy(cptr,varptr,sizeof(xmpVariable_t));
         strcpy(varptr->varName.key,rowKey);

         switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {

	   case 1: /* moduleName */
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            m->modulename);
	      break;

	   case 2: /* moduleVersion */
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            m->version);
	      break;

  	   case 3: /* moduleOwner */
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            m->owner);
	      break;

  	   case 4: /* moduleDependList */
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            m->dependList);
	      break;

   	   case 5: /* moduleType */
              varptr->varSyntax = integer;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                            m->modType);
	      break;

   	   case 6: /* moduleContent fetch/encode file/contents */
	      varptr->varSyntax = displayString;
              varptr->valueBuf = getModuleContents(m->modulename,
                                                   &varptr->valueStrLen);
              if (varptr->valueBuf == NULL) {
                 out->errorStatus = resourceUnavailable;
                 out->errorIndex = numVariables;
                 out->numberVariables = numVariables;
                 return resourceUnavailable;
              }
              break;

  	   case 7:  /* moduleTime */
	      xmpConvertTimeStr(m->timeStamp,timestr,sizeof(timestr));
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",timestr);
              break;

  	   case 8: /* pending version */
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            m->pendingVersion);
              break;

	   default:
	      fprintf(stderr,"var_modList: invalid column\n");
              /* release lock */
              pthread_mutex_unlock(&moduleList->lock);
              out->errorStatus = noSuchName;
              out->errorIndex = numVariables;
              out->numberVariables = numVariables;
              return noSuchName;

         } /* switch column */

         /* glue into variable list */
         numVariables++;
         if (out->varListTail)
            out->varListTail->next = varptr;
         out->varListTail = varptr;
         if (out->varList == NULL)
            out->varList = varptr;           

         cptr = cptr->next;

       } /* while columns */

       numRows++;
       e = e->next;

     } /* while modules */

     /* unlock */
     pthread_mutex_unlock(&moduleList->lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (xmpd_verbose)
        printf("var_modList: %d variables, finished\n",numVariables);

     return noError;
}

/* ************************************************************* */
/* look for special version string inside binary and return
   value or 0.0 if not found; if specialStr == NULL,
   then look for "KorgVersion = %3.2f"

   for now, only look for KorgVersion
*/

float findBinaryVersion(char *filename, char *specialStr)
{
     u_char *fileBuf;
     int fileBufSize,i;
     char version[12],*cptr;
     float v;
 
     if (filename == NULL)
        return 0.0;

     fileBufSize = 0;
     if ((fileBuf = read_file(filename,&fileBufSize)) == NULL) {
        return 0.0;
     }

     v = 0.0;
     for (i=0; i<fileBufSize; i++) {

         if (fileBuf[i] == (u_char)'K') {

	    if (strncmp((char *)&fileBuf[i],"KorgVersion ",12) == 0) {

	       /* version is the next %3.2f after KorgVersion */
	       cptr = (char *)&fileBuf[i+12];
               strncpy(version,cptr,11);
               version[11] = 0;
               v = (float)atof(version);
               /* keep checking so we dont find a bogus copy of KorgVersion */
               if (v != 0.0)
  	          break;
	    }

         }
     }

     if (fileBuf != NULL)
        free(fileBuf);

     return v;
}

/* ************************************************************* */

int init_modules()
{
     char version[64];
     float vers;

     if (autoUpgrade == 0) {
        return 0;
     }

     moduleList = newLinkedList(sizeof(module_t),compareModules,
                                printModule,freeModule,"coreModules",1);

     init = 1;

     /* seed it with base agent files/modules */

     /* xmpd binary -- xmpd-platform.suffix */

     /* regardless of the platform, we need to drag around the
        win32 dlls; we assume all agent binaries are of same
        version */

     vers = findBinaryVersion("xmpd-solaris",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("xmpd-solaris",version,"xmpd",NULL,agentModule);

     vers = findBinaryVersion("xmpd-linux",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("xmpd-linux",version,"xmpd",NULL,agentModule);

     vers = findBinaryVersion("xmpd-solx86",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("xmpd-solx86",version,"xmpd",NULL,agentModule);

     vers = findBinaryVersion("xmpd-win32.exe",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("xmpd-win32.exe",version,"xmpd","iconv.dll,libxml2.dll,pthreadVC.dll1,zlib1.dll",agentModule);

     vers = findBinaryVersion("cartographertray.exe",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("cartographertray.exe",version,"xmpd","iconv.dll,libxml2.dll,pthreadVC1.dll,zlib1.dll",agentModule);

     vers = findBinaryVersion("restartxmpd.exe",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("restartxmpd.exe",version,"xmpd",NULL,agentModule);

     vers = findBinaryVersion("restartxmpd.sh",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("restartxmpd.sh",version,"xmpd",NULL,agentModule);

     /* xmpd.xml */
     sprintf(version,"%d",xmpd_config_version);
     addModule("xmpd.xml",version,"xmpd",NULL,dataModule);

     /* win32-specific dlls we drag around */
     /* pthreadVC1.dll, libxml2.dll, iconv.dll, zlib1.dll */
     /* XXX we need to dynamically figure out versions */

     vers = findBinaryVersion("ntsetup.exe",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("ntsetup.exe",version,"xmpd",NULL,dataModule);

     /* XXX -- we dont know these versions yet but we need them */
     addModule("pthreadVC1.dll","1","xmpd",NULL,libModule);
     addModule("libxml2.dll","1","xmpd",NULL,libModule);
     addModule("iconv.dll","1","xmpd",NULL,libModule);
     addModule("zlib1.dll","1","xmpd",NULL,libModule);

     addModule("pcre.dll","8.01","xmpd",NULL,libModule);
     addModule("libpcre-sparcv9.so","8.01","xmpd",NULL,libModule);
     addModule("libpcre-linux.so","8.01","xmpd",NULL,libModule);
     addModule("libpcre-solx86.so","8.01","xmpd",NULL,libModule);

     /* SSL related libs that we drag around (all three platforms) */
     /* we dont know their versions per se but we also need them */
     addModule("libcrypto-linux.so","1","xmpd",NULL,libModule);
     addModule("libcrypto-sparcv9.so","1","xmpd",NULL,libModule);
     addModule("libcrypto-solx86.so","1","xmpd",NULL,libModule);
     addModule("libssl-linux.so","1","xmpd",NULL,libModule);
     addModule("libssl-sparcv9.so","1","xmpd",NULL,libModule);
     addModule("libssl-solx86.so","1","xmpd",NULL,libModule);
     addModule("libeay32.dll","1","xmpd",NULL,libModule);
     addModule("ssleay32.dll","1","xmpd",NULL,libModule);

     vers = findBinaryVersion("xmpquery-solaris",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("xmpquery-solaris",version,"xmpd",NULL,binModule);

     vers = findBinaryVersion("xmpquery-linux",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("xmpquery-linux",version,"xmpd",NULL,binModule);

     vers = findBinaryVersion("xmpquery-solx86",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("xmpquery-solx86",version,"xmpd",NULL,binModule);

     vers = findBinaryVersion("xmpquery-win32.exe",NULL);
     sprintf(version,"%3.2f",vers);
     addModule("xmpquery-win32.exe",version,"xmpd",NULL,binModule);

     return 1;
}

/* ************************************************************* */
/* given an established session, fetch its module table and return
   as an array; caller must free the table when done 
*/

module_t *getPeerModuleList(xmpSession_t *sptr, int *numModules)
{
     xmpPdu_t query, reply;
     xmpVariable_t var1,var2,var3, *varptr;
     int ret;
     module_t *moduleTable = NULL;
     int numPeerModules = 0;
     int i;
     char addrstr[64];

     if ((sptr == NULL) || (numModules == NULL)) 
        return NULL;

     *numModules = 0;

     if (xmpIsSessionConnected(sptr) == 0) {
        return NULL;
     }

     bzero((void *)&query,sizeof(query));

     query.msgType = SelectTableRequest;
     query.messageID = (int64_t)time(NULL);
     query.errorStatus = noError;
     query.errorIndex = 0;

     sprintf(query.tableMib,"core");
     sprintf(query.tableName,"moduleTable");
     sprintf(query.tableKey,"*");
     query.maxRows = 0;

     /* query for moduleName, moduleVersion */
     bzero((void *)&var1,sizeof(var1));     
     sprintf(var1.varName.mibName,"core");
     sprintf(var1.varName.objName,"moduleName");
     var1.varSyntax = nullSyntax;
     var1.valueStrLen = 0;
     var1.next = &var2;

     bzero((void *)&var2,sizeof(var2));
     sprintf(var2.varName.mibName,"core");
     sprintf(var2.varName.objName,"moduleVersion");
     var2.varSyntax = nullSyntax;
     var2.valueStrLen = 0;
     var2.next = &var3;

     bzero((void *)&var3,sizeof(var3));
     sprintf(var3.varName.mibName,"core");
     sprintf(var3.varName.objName,"moduleType");
     var3.varSyntax = nullSyntax;
     var3.valueStrLen = 0;
     var3.next = NULL;

     /* get module type as well in case peer has modules that we dont;
        we need to know how to deal with them */

     query.numberVariables = 3;
     query.varList = &var1;
     query.varListTail = &var3;

     bzero((void *)&reply,sizeof(reply));
     if ((ret = xmpSendQueryReceiveReply(sptr,&query,&reply)) < 0) {
        xmpConvertIpAddr(sptr->remIp,addrstr,63);
        fprintf(stderr,"getPeerModuleList: query %s failed\n",addrstr);
     }
     else {
        ;
     }

     moduleTable = NULL;
     *numModules = 0;

     if ((reply.errorStatus == noError) && (reply.numberVariables >= 3)) {

        numPeerModules = reply.numberVariables / 3;
        moduleTable = (module_t *)malloc(numPeerModules*sizeof(module_t));
        if (moduleTable == NULL) {
	   fprintf(stderr,"getPeerModuleList: malloc failed\n");
           xmpPduCleanup(&reply);
           return NULL;
        }

        bzero((void *)moduleTable,sizeof(module_t)*numPeerModules);

        varptr = reply.varList;
        for (i=0; i<numPeerModules; i++) {

	    if (varptr != NULL) {
	       strncpy(moduleTable[i].modulename,varptr->valueStr,255);
               moduleTable[i].modulename[255] = 0;
               varptr = varptr->next;
            }
            if (varptr != NULL) {
	       strncpy(moduleTable[i].version,varptr->valueStr,255);
               moduleTable[i].version[255] = 0;
               varptr = varptr->next;
            }
            if (varptr != NULL) {
	       moduleTable[i].modType = atoi(varptr->valueStr);
               varptr = varptr->next;
            }

        }

        *numModules = numPeerModules;
     }

     xmpPduCleanup(&reply);

     return moduleTable;
}

/* ************************************************************* */
/* given a peer's module table, compare and see if there is some newer 
   version I could retrieve from this peer
 
   this function is where the magic occurs -- for each of our
   modules, we see if its in peer's list and if there version is 
   greater than ours.

   if the peer has files that I do not, I should also retrieve
   them as well
*/

int compareModuleLists(char *peerName, xmpSession_t *session,
                       module_t *peerModuleTable, 
                       int numPeerModules)
{
     module_t *myModules = NULL;
     int numMyModules = 0;
     int i,j;
     double f1, f2, f3;
     int update = 0;
     char messageStr[512];

     if ((peerName == NULL) || (peerModuleTable == NULL) || 
         (numPeerModules < 1)) {
        return -1;
     }

     /* get a local copy of our own modules */
     numMyModules = 0;
     myModules = (module_t *)toArrayCopy(moduleList,&numMyModules);

     /* if I dont have anything to compare, dont bother */
     if ((myModules == NULL) || (numMyModules < 1)) {
        return -1;
     }

     /* for each of my modules, see if peer has it and if its newer
        (i.e. has greater version number) */
     for (i=0; i<numMyModules; i++) {

         f1 = atof(myModules[i].version);

         if (f1 == (double)0.0) {
	    DBG1("compareModuleLists: not bothering to upgrade %s because its experimental\n",myModules[i].modulename);
            continue; /* with for i=0 loop */
         }

         f3 = 0.0;
         if (myModules[i].pendingVersion[0] != 0)
	    f3 = atof(myModules[i].pendingVersion);

         for (j=0; j<numPeerModules; j++) {

	     if (strcmp(myModules[i].modulename,
                        peerModuleTable[j].modulename) == 0) {

	        f2 = atof(peerModuleTable[j].version);

                /* fprintf(stderr,"Matched '%s,%d',%3.2f/%3.2f\n",
                        myModules[i].modulename,myModules[i].modType,
                        f1,f2);
		*/

                if ((f2 > f1) && (f2 > f3)) {
	  	   fprintf(stderr,"%s has new version of %s, %3.2f %3.2f\n",
                           peerName,myModules[i].modulename,f2,f1);

                   /* go fetch the module and then arrange to
                      install it via restart 
                     */

                   if (fetchModule(peerName,session,myModules[i].modulename,
                                   myModules[i].modType,
                                   peerModuleTable[j].version) > 0) {
		      sprintf(myModules[i].pendingVersion,"%3.2f",f2);
                      sprintf(messageStr,"Downloaded %s(%3.2f) from %s",
                              myModules[i].modulename,
                              f2,
                              peerName);
                      xmpd_log(messageStr);
	  	      update = 1;

                      fprintf(stderr,"%s\n",messageStr);

                      /* unfortunately, writing to myModules[] does
                         not update the main module list though */
                      updateModulePendingVersion(myModules[i].modulename,
                                                 f2);
                      
                   } /* if fetchModule() */

                } /* if peer has newer module than mine */
                
             } /* if modules match */

         } /* for each peer module */

     } /* for each of my own modules */

     /* for each of peer's modules, see if he's got one that I dont
        have */
     for (j=0; j<numPeerModules; j++) {

         f1 = atof(peerModuleTable[j].version);
         for (i=0; i<numMyModules; i++) {
	     if (strcmp(peerModuleTable[j].modulename,
                        myModules[i].modulename) == 0)
	         break; /* from for loop */
         }
         if (i == numMyModules) {
 	    fprintf(stderr,"%s has module '%s' that I dont have\n",
                    peerName,peerModuleTable[j].modulename);

             /* go fetch the module and then arrange to
                install it either by restart or simply rename XXX
             */

	    if (fetchModule(peerName,session,peerModuleTable[j].modulename,
                            peerModuleTable[j].modType,
                            peerModuleTable[j].version) > 0) {

                sprintf(messageStr,"Downloaded new %s(%s) from %s",
                        peerModuleTable[j].modulename,
                        peerModuleTable[j].version,
                        peerName);
                xmpd_log(messageStr);

                fprintf(stderr,"%s\n",messageStr);

                /* should we add this new module to our own core modules
                   list ? XXX */

	        update  = 1;
	    }

         } /* if I did not find a match for peer module */

     } /* for each peer module */

     /* cleanup, done */
     if (myModules != NULL)
        free(myModules);

     return update;
}

/* ************************************************************* */
/* go through updated files in our working-directory
   and figure out what to do with them

   Files will have format of name.new where name
   most likely matches an existing module
*/

#ifndef WIN32

/* UNIX */

/* fork off and detach restartxmpd.sh */
int restartCartographer()
{
    char commandline[256];
    char commandArg[256];
    char *commandArgv[3];
    pid_t pid;
    struct sigaction sa;
    int i;

    /* set signal for no child waiting; we do this in main.c but
       do it again just to be darn sure */
    sa.sa_handler = SIG_IGN;
    sa.sa_flags = SA_NOCLDSTOP;
    sigemptyset(&sa.sa_mask);
    sigaction(SIGCHLD,&sa,NULL);

    /* fork */
    pid = fork();

    if (pid == 0) {
       /* child process - exec the command passing the homedir
          as the first/only argument */

       /* before we exec, we want to make sure to close all file descriptors
          so we dont hold open sockets; in testing, we found that
          our shell script held open our main socket mucking up restarting */
       for (i=3; i<NOFILE; i++)
	   close(i);

       sprintf(commandline,"%s/restartxmpd.sh",xmpd_home_dir);

       /* construct arg array */
       commandArgv[0] = commandline;
       commandArgv[1] = commandArg;
       commandArgv[2] = NULL;
       sprintf(commandArg,"%s",xmpd_home_dir);
       execv(commandline,commandArgv);

       /* if exec fails, we end up here */
       xmpd_log("Failed to exec restartxmpd.sh\n");
       exit(1);
    }
    else {
       /* parent; dont do anything really */
       if (pid < 0) {
	  xmpd_log("Failed to fork to start restartxmpd.sh\n");
       }
    }

    return 1;
}

/* deprecated */
int parseDirForUpdates() 
{
     DIR *currentdir;
     struct dirent *dent,*dptr;
     u_char dbuf[sizeof(struct dirent)+FILENAME_MAX+1];
     module_t *m,comp;
     int restart = 0;
     char moduleName[256],*cptr, oldModuleName[512];
     linkedElement_t *e;
     float v;

     if ((currentdir = opendir(".")) == NULL) {
        return -1;
     }

     /* look for blah.new */
     dent = (struct dirent *)dbuf;
     while (readdir_r(currentdir,dent,&dptr) == 0) {

       /* sanity checks */
       if (dptr == NULL) {
	  break;
       }

       if (strstr(dent->d_name,".new") == NULL)
	  continue;

       /* find the existing module's entry (w/o '.new') */
       strncpy(moduleName,dent->d_name,255);
       moduleName[255] = 0;
       strncpy(comp.modulename,dent->d_name,255);
       comp.modulename[255] = 0;
       cptr = strstr(comp.modulename,".new");
       if (cptr != NULL)
	  *cptr = 0;
       e = findElement(moduleList,(void *)&comp);
       if (e == NULL) {
 	  fprintf(stderr,"parseDirForUpdates: new unknown module %s\n",
                  moduleName);

          /* rename blah.new to simply blah */
          rename(moduleName,comp.modulename);
	  continue;
       }
 
       m = (module_t *)e->data;
       if (m == NULL)
	  continue;

       /* if its data or bin module, rename old, then rename new */
       switch (m->modType) {

         case agentModule:
         case pluginModule:
         case libModule:
            /* if its lib, agent, or plugin, restart agent once we've 
               finished the loop */
	    restart = 1;
	    break;

         case dataModule:
         case binModule:
	    /* rename blah to blah.old */
	    sprintf(oldModuleName,"%s.old",comp.modulename);
	    rename(comp.modulename,oldModuleName);

            /* rename blah.new to blah */
            rename(dent->d_name,m->modulename);

            /* if binModule, update entry with new version */
            if (m->modType == binModule) {
	       v = findBinaryVersion(m->modulename,NULL);
               sprintf(m->version,"%3.2f",v);
            }

	    /* if data module, the plugin or entity that re-reads
               the datafile will update its entry */

	    break;

         invalidModuleType:
         default:
	    fprintf(stderr,"parseDirForUpdates: unknown mod type %d for %s\n",
		    m->modType,moduleName);
       }

     } /* while dir entries */ 

     /* dont forget to close dir!!! */
     closedir(currentdir);

     if (restart) {
        /* fork off restart program XXX */

     }

     return 1;
} /* parseDirForUpdates() */

#else

/* WIN32 */

/* fork off and detach restartxmpd.sh */
int restartCartographer()
{
     char commandline[256];
     PROCESS_INFORMATION pinfo;
     STARTUPINFO sinfo;
     int bool;

     /* set up the command-line and args */
     sprintf(commandline,"%s\\restartxmpd.exe %s",
             xmpd_home_dir,
             xmpd_home_dir);
     bzero(&pinfo,sizeof(pinfo));
     bzero(&sinfo,sizeof(sinfo));
     sinfo.cb = sizeof(sinfo);

     /* use CreateProcess() to start process */
     bool = CreateProcess(NULL, /* app-name */
                          commandline, 
                          NULL, /* process attributes */
                          NULL, /* thread attributes */
                          FALSE, /* false=dont inherit handles */
                          DETACHED_PROCESS, /* creation flags */
                          NULL, /* environment, use parent's */
                          NULL, /* use cwd of cartographer */
                          &sinfo,
                          &pinfo);
     
     if (bool == 0) {
        xmpd_log("Calling restartxmpd.exe failed");
        return -1;
     }
     else {
        xmpd_log("Calling restartxmpd.exe succeeded");
     }

     /* dont forget to close handles */
     CloseHandle(pinfo.hProcess);
     CloseHandle(pinfo.hThread);

     return 1;
}

/* deprecated */
int parseDirForUpdates() 
{
     HANDLE hFind;
     WIN32_FIND_DATA dent;
     char searchdir[1024];
     module_t *m,comp;
     int restart = 0;
     char moduleName[256],*cptr, oldModuleName[512];
     linkedElement_t *e;
     float v;

     sprintf(searchdir,"*.new");

     hFind = FindFirstFile(searchdir,&dent);

     /* look for *.new files */
     while (hFind != INVALID_HANDLE_VALUE) {

       if (strstr(dent.cFileName,".new") == NULL) {
          if (FindNextFile(hFind,&dent) == 0) {
   	     break;
          }
	  continue;
       }

       /* find the existing module's entry (w/o '.new') */
       strncpy(moduleName,dent.cFileName,255);
       moduleName[255] = 0;
       strncpy(comp.modulename,dent.cFileName,255);
       comp.modulename[255] = 0;
       cptr = strstr(comp.modulename,".new");
       if (cptr != NULL)
	  *cptr = 0;
       e = findElement(moduleList,(void *)&comp);
       if (e == NULL) {
 	  fprintf(stderr,"parseDirForUpdates: new unknown module %s\n",
                  moduleName);

          /* rename blah.new to simply blah */
          rename(moduleName,comp.modulename);
          if (FindNextFile(hFind,&dent) == 0) {
   	     break;
          }
	  continue;
       }
 
       m = (module_t *)e->data;
       if (m == NULL) {
          if (FindNextFile(hFind,&dent) == 0) {
   	     break;
          }
	  continue;
       }

       /* if its data or bin module, rename old, then rename new */
       switch (m->modType) {

         case agentModule:
         case pluginModule:
         case libModule:
            /* if its lib, agent, or plugin, restart agent once we've 
               finished the loop */
	    restart = 1;
	    break;

         case dataModule:
         case binModule:
	    /* rename blah to blah.old */
	    sprintf(oldModuleName,"%s.old",comp.modulename);
	    rename(comp.modulename,oldModuleName);

            /* rename blah.new to blah */
            rename(dent.cFileName,m->modulename);

            /* if binModule, update entry with new version */
            if (m->modType == binModule) {
	       v = findBinaryVersion(m->modulename,NULL);
               sprintf(m->version,"%3.2f",v);
            }

	    /* if data module, the plugin or entity that re-reads
               the datafile will update its entry */

	    break;

         invalidModuleType:
         default:
	    fprintf(stderr,"parseDirForUpdates: unknown mod type %d for %s\n",
		    m->modType,moduleName);
       }


       if (FindNextFile(hFind,&dent) == 0) {
	  break;
       }

     } /* while */

     FindClose(hFind); 

     return 1;
}

#endif /* !WIN32 */

/* ************************************************************* */
/* check software modules of peer and pull any necessary
   downloads/upgrades and then install them */

int checkPeerModules(char *peerName)
{
     u_int ipv4addr;
     xmpSession_t *session;
     module_t *peerModuleList;
     int numPeerModules,ret;
     char vers[128],sysDescr[128];

     if (autoUpgrade != 1) { 
        DBG("checkPeerModules: autoUpgrade=0\n");
        return 0;
     }
     if (upgradeNeeded != 0) {
        /* if we have already downloaded some upgrades, 
           dont keep downloading as we may be terminated 
           in the middle of a download; wait to get upgrades
           after we re-start */
        DBG("checkPeerModules: upgradeNeeded=1\n");
        return 0;
     }

     if (peerName == NULL) {
         DBG("checkPeerModules: null peerName\n");
         return -1;
     }

     /* disable peer modules? If switch is set */
     if (xmpd_no_modules) {
        DBG("checkPeerModules: xmpd_no_modules=1\n");
        return 0;
     }

     DBG1("checkPeerModules: checking %s\n",peerName);

     /* establish session with peer */
     ipv4addr = xmpHostnameToIpaddr(peerName);

     if ((session = newXmpSessionConnectTimeout(ipv4addr,
			(u_short)htons((u_short)XMP_PORT),3,NULL)) == NULL) {
        return 0;
     }
     if (xmpGetPeerVersion(session,vers,sysDescr,127) < 0) {
        destroyXmpSession(session);
        return 0;
     }

     /* get peer's module list */
     peerModuleList = getPeerModuleList(session,&numPeerModules);

     ret = compareModuleLists(peerName,session,peerModuleList,numPeerModules);

     if (ret > 0) {

        /* we've obtained updates; regardless of what they are,
           lets schedule a brain transplate and restart ourselves
           after we give ourselves some time to settle down and
           check a few other systems first */

        DBG("checkPeerModules: upgrade needed\n");

        upgradeNeeded = 1;
        upgradeNeededTime = time(NULL);
     }
 
     /* done, do cleanup */
     destroyXmpSession(session);

     /* cleanup peerModuleList since we essentially malloc'd it */
     if (peerModuleList != NULL)
        free(peerModuleList);

     return 1;
}

/* ************************************************************* */
/* called periodically by housekeeping to initiate restart
   of Cartographer agent along with upgrade */

int upgradeCartographer()
{
     static int upgradeLaunched = 0;

     if (upgradeNeeded == 0) {
        return 0;
     }
     if (autoUpgrade == 0)
        return 0;

     /* if an upgraded is needed, but we have already 
        launched the script/program, then just return.
        We want to leave the global flag set so that checkPeers()
        will not continue make connections during shutdown */

     if (upgradeLaunched)
        return 0;

     /* we want to wait 5 minutes or so before doing
        an upgrade */
     if ((time(NULL) - upgradeNeededTime) < 300) {
        return 0;
     }

     /* launch restartxmpd.exe or restartxmpd.sh */
     fprintf(stderr,"updateCartographer: going to restart Cartographer\n");
     restartCartographer();

     /* now that we have launched the daemon, set the flag  
        so that if we are called again whilest still waiting
        to exit, we'll not re-run restartxmpd.sh/.exe */
    
     upgradeLaunched = 1;

     return 1;
}

/* ************************************************************* */

