/* 
   COPYRIGHT 2010 KRUPCZAK.ORG, LLC.

   THIS SOFTWARE IS THE CONFIDENTIAL AND PROPRIETARY PRODUCT OF 
   KRUPCZAK.ORG, LLC.  ANY UNAUTHORIZED USE, REPRODUCTION, OR
   TRANSFER OF THIS SOFTWARE IS STRICTLY PROHIBITED.  (SUBJECT TO LIMITED
   DISTRIBUTION AND RESTRICTED DISCLOSURE ONLY.)  ALL RIGHTS RESERVED.
*/

/* file - linkedlist.c */
/* rdk@krupczak.org */

/* code for generic doubly linked list package */

/* add support for auto lock or not locking */

#ifdef WIN32
#include <windows.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#include <strings.h>
#endif

#include "linkedlist.h"

/* ************************************************************* */
/* do we want to store a copy of lists that we malloc ? XXX */

linkedList_t *newLinkedList(int dataSize, void *compareptr, void *printptr,
                            void *freeptr, char *descr, int autoLock)
{
     linkedList_t *list;

     if ((list = (linkedList_t *)malloc(sizeof(linkedList_t))) == NULL) {
#ifndef WIN32
        perror("malloc of linked list failed");
#else
        fprintf(stderr,"malloc of linked list failed, %d\n",GetLastError());
#endif
        return NULL;
     }

     list->dataSize = dataSize;
     list->head = list->tail = NULL;
     list->elementCompare = compareptr;
     list->elementPrint = printptr;
     list->elementFree = freeptr;

     list->autoLock = autoLock; /* 0,1 for lock/nolock */

     list->description[0] = 0;
     if (descr != NULL) {
        strncpy(list->description,descr,63);
        list->description[63] = 0;
     }

     /* init the lock just in case ? */
     pthread_mutex_init(&list->lock,NULL);

     return list;
}

/* ************************************************************* */
/* destroy linked list and any element on it; call user's
   free element for their data portion, if that function-pointer
   is defined; otherwise, we do not free user-defined data 
*/

int destroyLinkedList(linkedList_t *list)
{
     linkedElement_t *e,*t;

     if (list == NULL) {
        return -1;
     }

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     e = list->head;

     while (e != NULL) {
       if (list->elementFree != NULL)
 	  (list->elementFree)(e->data);
       t = e;
       e = e->next;
       free(t);
     }

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     /* destroy lock regardless */
     pthread_mutex_destroy(&list->lock);

     free(list);

     return 0;
}

/* ************************************************************* */
/* given a linked list, destroy each element and any
   user-defined data if applicable; leave the list intact
   and empty 
*/

int drainLinkedList(linkedList_t *list)
{
     linkedElement_t *e,*t;

     if (list == NULL) {
        return -1;
     }

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     e = list->head;

     while (e != NULL) {

       if ((list->elementFree != NULL) && (e->data != NULL))
 	  (list->elementFree)(e->data);
       t = e;
       e = e->next;
       free(t);
     }

     list->head = list->tail = NULL;

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     return 0;
}

/* ************************************************************* */

int insertInOrder(linkedList_t *list, void *data)
{
     int ret;
     linkedElement_t *e, *ptr;

     if ((list == NULL) || (data == NULL)) {
        return -1;
     }

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     /* if list does not have a user-defined compare function,
        simply insert on head */

     if (list->elementCompare == NULL) {
        pthread_mutex_unlock(&list->lock);
        return insertHead(list,data);
     }

     /* malloc a new element */
     if ((e = (linkedElement_t *)malloc(sizeof(linkedElement_t))) == NULL) {
        if (list->autoLock)
           pthread_mutex_unlock(&list->lock);
        return -1;
     }

     e->data = data;
     e->next = e->prev = NULL;

     ptr = list->head;
     while (ptr != NULL) {

       ret = (list->elementCompare)(data,ptr->data);

       if (ret < 0)
	  break;

       ptr = ptr->next;
     }     

     if (ptr == NULL) {
        /* add to tail */
        if (list->autoLock)
           pthread_mutex_unlock(&list->lock);
        return insertTail(list,data);
     }
     else {
        /* add before ptr since ptr is first object greater  */

        e->next = ptr;
        e->prev = ptr->prev;
        ptr->prev = e;

        /* ptr->next unchanced */

        if (e->prev != NULL)
 	   e->prev->next = e;
        if (list->head == ptr)
 	   list->head = e;
     }

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);
     
     return 0;
}

/* ************************************************************* */

int insertHead(linkedList_t *list, void *data)
{
     linkedElement_t *e;

     if ((list == NULL) || (data == NULL)) {
        return -1;
     }

     /* malloc a new element */
     if ((e = (linkedElement_t *)malloc(sizeof(linkedElement_t))) == NULL) {
        return -1;
     }

     e->data = data;
     e->next = e->prev = NULL;

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     /* initial case */
     if (list->head == NULL) {
        list->head = e;
        if (list->tail == NULL) {
 	   list->tail = e;
        }
     }
     else {
        e->next = list->head;
        list->head->prev = e;
        list->head = e;
     }

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     return 0;
}

/* ************************************************************* */

int insertTail(linkedList_t *list, void *data)
{
     linkedElement_t *e;

     if ((list == NULL) || (data == NULL)) {
        return -1;
     }

     /* malloc a new element */
     if ((e = (linkedElement_t *)malloc(sizeof(linkedElement_t))) == NULL) {
        return -1;
     }

     e->data = data;
     e->next = e->prev = NULL;

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     /* initial case */
     if (list->tail == NULL) {
        list->tail = e;
        if (list->head == NULL) {
  	   list->head = e;
        }
     }
     else {
        list->tail->next = e;
        e->next = NULL;
        e->prev = list->tail;
        list->tail = e;
     }

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     return 0;
}

/* ************************************************************* */
/* find an element; use user-supplied find method and a pointer
   to some opaque data that user provides; we then call user's
   compare function; if ret=0, we return that element still
   in list 
*/

linkedElement_t *findElement(linkedList_t *list, void *comparedata)
{
     linkedElement_t *e;
     int ret;

     if ((list == NULL) || (comparedata == NULL)) {
        return NULL;
     }

     if (list->autoLock)
         pthread_mutex_lock(&list->lock);

     /* if list does not have a user-defined compare function,
        we cant proceed */

     if (list->elementCompare == NULL) {
        if (list->autoLock)
           pthread_mutex_unlock(&list->lock);
        return NULL;
     }

     e = list->head;

     while (e != NULL) {
       ret = (list->elementCompare)(e->data,comparedata);
       if (ret == 0)
	  break;

       e = e->next;
     }

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     return e;
}

/* ************************************************************* */
/* get number elements */

int numberElements(linkedList_t *list)
{
     linkedElement_t *ptr;
     int number = 0;

     if (list == NULL) {
        return 0;
     }

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     number = 0;
     ptr = list->head;
     while (ptr != NULL) {
       number++;
       ptr = ptr->next;
     }

     if (list->autoLock)
         pthread_mutex_unlock(&list->lock);

     return number;
}

/* ************************************************************* */
/* remove element e from list; caller is traversing list
   and already has lock (if autoLock) */

int removeElement(linkedList_t *list, linkedElement_t *e)
{
     linkedElement_t *ptr;

     if (list == NULL) {
        return -1;
     }
     if (e == NULL)
        return -1;    

     ptr = e;

     /* ptr is now what we want to destroy */
 
     if (ptr->prev != NULL) {
        ptr->prev->next = ptr->next;
     }
     if (ptr->next != NULL) {
        ptr->next->prev = ptr->prev;
     }
     if (ptr == list->head) {
        list->head = ptr->next;
     }
     if (ptr == list->tail) {
        list->tail = ptr->prev;
     }

     ptr->next = ptr->prev = NULL;

     /* now free ptr */
     if (list->elementFree != NULL) {
        (list->elementFree)(ptr->data);
     }
     free(ptr);

     return 1;
}

/* ************************************************************* */
/* remove the element from the list (disconnect), free the linked element,
   but *do not* free the data portion; caller is traversing the list
   and already has the lock (if autoLock) */

int disconnectElement(linkedList_t *list, linkedElement_t *e)
{
     linkedElement_t *ptr;

     if (list == NULL) {
        return -1;
     }
     if (e == NULL)
        return -1;    

     ptr = e;

     /* ptr is now what we want to destroy */
 
     if (ptr->prev != NULL) {
        ptr->prev->next = ptr->next;
     }
     if (ptr->next != NULL) {
        ptr->next->prev = ptr->prev;
     }
     if (ptr == list->head) {
        list->head = ptr->next;
     }
     if (ptr == list->tail) {
        list->tail = ptr->prev;
     }

     ptr->next = ptr->prev = NULL;

     /* now free ptr, but not data */
     free(ptr);

     return 1;
}

/* ************************************************************* */
/* destroy element e which is somewhere in the list */
/* find element first */

int destroyElement(linkedList_t *list, linkedElement_t *e)
{
     linkedElement_t *ptr;

     if (list == NULL) {
        return -1;
     }

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     /* find e in linked list */
     ptr = list->head;
     while (ptr != NULL) {
       if (ptr == e)
	  break;
       ptr = ptr->next;
     }
     if (ptr == NULL) {
        fprintf(stderr,"Couldnt find element %lx in list %lx\n",
                e,list);
        if (list->autoLock)
           pthread_mutex_unlock(&list->lock);
        return -1;
     }

     /* ptr is now what we want to destroy */
 
     if (ptr->prev != NULL) {
        ptr->prev->next = ptr->next;
     }
     if (ptr->next != NULL) {
        ptr->next->prev = ptr->prev;
     }
     if (ptr == list->head) {
        list->head = ptr->next;
     }
     if (ptr == list->tail) {
        list->tail = ptr->prev;
     }

     ptr->next = ptr->prev = NULL;

     /* now free ptr */
     if (list->elementFree != NULL) {
        (list->elementFree)(ptr->data);
     }
     free(ptr);

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     return 0;
}

/* ************************************************************* */

void *popHead(linkedList_t *list) 
{ 
     linkedElement_t *e;
     void *ptr;

     if (list == NULL) {
        return NULL;
     }

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     /* the element to return */
     if ((e = list->head) == NULL) {
        if (list->autoLock)
           pthread_mutex_unlock(&list->lock);
        return NULL;
     }

     if (e->prev != NULL)
        e->prev->next = e->next;
     if (e->next != NULL)
        e->next->prev = e->prev;
     if (e == list->head)
        list->head = e->next;
     if (e == list->tail)
        list->tail = e->prev;

     e->next = e->prev = NULL;

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     ptr = e->data;
     e->data = NULL;
     free(e);
     
     return ptr; 
}

/* ************************************************************* */

void *popTail(linkedList_t *list) 
{ 
     linkedElement_t *e;
     void *ptr;

     if (list == NULL) {
        return NULL;
     }

     /* get lock */
     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     if ((e = list->tail) == NULL) {
        if (list->autoLock)
           pthread_mutex_unlock(&list->lock);
        return NULL;
     }

     if (e->prev != NULL)
        e->prev->next = e->next;
     if (e->next != NULL)
        e->next->prev = e->prev;
     if (e == list->head)
        list->head = e->next;
     if (e == list->tail)
        list->tail = e->prev;

     e->next = e->prev = NULL;

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     ptr = e->data;
     e->data = NULL;
     free(e);

     return ptr; 
}

/* ************************************************************* */
/* 
   copy the data from the linked list into an array and return
   to user; user frees the data if necessary 
*/

u_char *toArrayCopy(linkedList_t *list, int *numEntries)
{
     u_char *copy;
     int totalSize,offset;
     int n;
     linkedElement_t *e;

     if ((list == NULL) || (numEntries == NULL))
        return NULL;

     *numEntries = 0;

     if (list->dataSize == 0)
        return NULL;

     n = numberElements(list);
     totalSize = list->dataSize * n;

     if ((copy = (void *)malloc(totalSize)) == NULL) {
        return NULL;
     }
     bzero((void *)copy,totalSize); /* valgrind happy? */
 
     /* get lock */
     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     e = list->head;
     offset = 0;
     while ((e != NULL) && (n > 0)) {

       bcopy(e->data,&copy[offset],list->dataSize);

       (*numEntries)++;
       offset += list->dataSize;

       e = e->next;
       n = n-1;
     }

     if (list->autoLock)          
        pthread_mutex_unlock(&list->lock);

     return copy;
}

/* ************************************************************* */

int traverseList(linkedList_t *list) 
{
     linkedElement_t *e;

     if (list == NULL) {
        return -1;
     }

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     fprintf(stderr,"list: %s %lx, datasize:%d\n",
             list->description,
             list,list->dataSize);
     fprintf(stderr,"head:%lx tail:%lx\n",list->head,list->tail);

     e = list->head;

     while (e != NULL) {

       fprintf(stderr,"Element %lx\n",e);
       if (list->elementPrint)
	  (list->elementPrint)(e->data);
       else
 	  fprintf(stderr,"\tData:%lx\n",e->data);

       e = e->next;
     }

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);

     return 0;
}

/* ************************************************************* */
/* given a list, iterate function fptr over each element data */

int iterateList(linkedList_t *list, int (*fptr)())
{
     linkedElement_t *e;

     if ((list == NULL) || (fptr == NULL))
        return -1;

     if (list->autoLock)
        pthread_mutex_lock(&list->lock);

     e = list->head;

     while (e != NULL) {

       (fptr)(e->data);

       e = e->next;
     }

     if (list->autoLock)
        pthread_mutex_unlock(&list->lock);   

     return 1;

}

/* ************************************************************* */
