/* 
   COPYRIGHT 2007 KRUPCZAK.ORG, LLC.


   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - plugins.c */
/* rdk@krupczak.org */

/* Plugin loading, unloading, etc. */

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <sys/stat.h>

#ifndef WIN32
#include <strings.h>
#include <errno.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <dlfcn.h>
#include <dirent.h>
#endif

#ifdef WIN32
#include <windows.h>
#endif

#include <stdlib.h>
#include <string.h>

#include "xmp.h"
#include "xmp_vars.h"
#include "plugins.h"
#include "globals.h"
#include "socklib.h"
#include "xmp_session.h"
#include "pluginexport.h"
#include "linkedlist.h"
#include "xmp_events.h"
#include "modules.h"

/* ************************************************************* */
/* 
   plugin linked list; plugins are not ordered in any particular manner
   as plugins are loaded, they are put on head of list; as they are
   unloaded, their structures are zero'd out and flags set to EMPTY 

   updated 4/23/2007 to use our linked list module thus flags are
   irrelevant
*/

/* 
   we are implementing a two-pass unload procedure -- first pass
   calls the plugin's finish process and records the time.
   30-seconds later, it is assumed that all threads and resources
   within the plugin are stopped, data files are written, etc.
   then, daemon comes and actually dlclose's it and sets flags
   to PLUGIN_EMPTY.  

   If trying to load a plugin and flags are PLUGIN_UNLOADING, wait for 
   plugin to finish unloading before re-loading it

   Need to test XXX
*/


#if 0
/* lock variables */
plugin_t *plugin_list = NULL;
pthread_mutex_t plugin_lock = PTHREAD_MUTEX_INITIALIZER;
#endif

static linkedList_t *pluginList = NULL;

/* ************************************************************* */
/* routines called from linked list code */

int pluginPrint(plugin_t *p)
{
     if (p) {
       fprintf(stderr,"Plugin: name:'%s' flags:%04x path:'%s' descr:'%s'\n",
               p->modulename,
               p->flags,
               p->pathname,
               p->descr);
     }

     return 0;
}

int pluginCompare(plugin_t *a, plugin_t *b)
{
     if ((a == NULL) || (b == NULL))
        return -1;

     return strcmp(a->modulename,b->modulename);
}

/* free a plugin struct; can be called from linked list routines
   or by loading/unloading functions; right now, we simply
   free the structure; we also assume that the plugin structure
   has already been removed from the linked list
 */
int pluginFree(plugin_t *p)
{
     if (p) {
        free(p);
     }
     return 0;
}

/* ************************************************************* */
/* traverse plugin list, print list of names */

int traverse_plugins()
{

     traverseList(pluginList);

#if 0
     plugin_t *ptr;

     /* get lock */
     pthread_mutex_lock(&plugin_lock);

     ptr = plugin_list;
     while (ptr != NULL) {
       if (ptr->flags != PLUGIN_EMPTY) {
	  printf("Plugin: flags:%x Name:%s\n",ptr->flags,ptr->modulename);
       }
       ptr = ptr->next;
     }

     /* unlock */
     pthread_mutex_unlock(&plugin_lock);

#endif

     return 1;    
}

/* ************************************************************* */
/* called during agent initialization */

int initPlugins()
{
     pluginList = newLinkedList(sizeof(plugin_t),
                                pluginCompare,
                                pluginPrint,
                                pluginFree,
                                "Plugin linked list",1);
     return 0;
}

/* ************************************************************* */
/* create/initialize a plugin; caller has to then insert into list */


static plugin_t *find_unused_plugin()
{
     plugin_t *ptr;

     if ((ptr = (plugin_t *)malloc(sizeof(plugin_t))) == NULL) {
        fprintf(stderr,"malloc of plugin failed, %d\n",errno);
        return NULL;
     }

     bzero((void *)ptr,sizeof(plugin_t));
     ptr->flags = PLUGIN_EMPTY;
     
     return ptr;
}

/* ************************************************************* */
/* first phase of plugin un-load */

static int unloadPluginByPtr(plugin_t *p)
{
     if (p == NULL) {
        return -1;
     }

     /* call finish routine to start the ball rolling */
     (p->finish)();

     p->flags = PLUGIN_UNLOADING;
     p->loadTime = time(NULL);

     printf("unloadPluginByPtr: plugin %s unloading\n",p->modulename);

     return 1;
}

/* ************************************************************* */
/* unload and un-initialize plugin; assume we have lock */

static int unloadPluginByPtrFinal(plugin_t *p)
{
     if (p == NULL) {
        return -1;
     }

     p->loadTime = 0;
#ifndef WIN32
     dlclose(p->solibhandle);
#else
     FreeLibrary(p->solibhandle);
#endif
     p->flags = PLUGIN_EMPTY;
     p->solibhandle = NULL;

     printf("unloadPluginByPtrFinal: plugin %s unloading\n",p->modulename);

     return 1;
}

/* ************************************************************* */
/* get a plugin matching filename; we assume lock is already held */

static plugin_t *getPluginByFilename(char *filename)
{
     plugin_t *p;
     linkedElement_t *e;

     if (filename == NULL) {
        return NULL;
     }

     pthread_mutex_lock(&pluginList->lock);

     e = pluginList->head;
     while (e != NULL) {
       p = (plugin_t *)e->data;
       if ((p) && (strcmp(p->pathname,filename) == 0)) {
          pthread_mutex_unlock(&pluginList->lock);
	  return p;
       }

       e = e->next;
     }

     pthread_mutex_unlock(&pluginList->lock);

     return NULL;

#if 0
     ptr = plugin_list;

     while (ptr != NULL) {
       if ((ptr->flags != PLUGIN_EMPTY) &&
           (strcmp(ptr->pathname,filename) == 0))
	  return ptr;
       ptr = ptr->next;
     }

     return NULL;
#endif
}

/* ************************************************************* */
/* load and initialize a plugin; called either by agent engine
   or ourselves */

#ifndef WIN32

static int loadPlugin(char *filename)
{
     plugin_t *ptr;
     int ret;
     struct stat statbuf;
     char *aPtr;

     if (filename == NULL) {
        return -1;
     }

     DBG1("loadPlugin: starting to load '%s'\n",filename);

     if (xmpd_verbose)
        printf("loadPlugin: checking %s\n",filename);

     /* check to see if plugin has already been added;
        if so, unload old one and load new one */
     if ((ptr = getPluginByFilename(filename)) != NULL) {

        if (ptr->flags & PLUGIN_UNLOADING) {
           if (xmpd_verbose)
           printf("loadPlugin: not loading %s (yet) since its unloading\n",
	          ptr->modulename);
	   return 0;
        }

        /* check to see if the file has been modified
           since we loaded it; if it has not, then
           dont bother re-loading it */

        if (stat(filename,&statbuf) < 0) {
 	   perror("loadPlugin: stat of already loaded plugin failed");
           return -1;
        }

        /* if has not been modified since loading, dont reload */
        if (statbuf.st_mtime < ptr->loadTime) {
	  if (xmpd_verbose)
	     printf("loadPlugin: %s has not been updated\n",filename);
 	   return 0;
        }

        if (ptr->flags & PLUGIN_VALID) {
	  if (xmpd_verbose)
	      printf("loadPlugin: going to unload %s\n",ptr->modulename);
           unloadPluginByPtr(ptr);
           return 0;
        }

        /* at this point, plugin is already unloaded, proceed */

     }
     else {
        ptr = find_unused_plugin();
     }

     /* now, ptr is unused/empty plugin struct */
     ptr->flags = PLUGIN_VALID;

     if (xmpd_verbose)
        printf("loadPlugin: calling dlopen\n");

     /* open shared library via dlopen */
     if ((ptr->solibhandle = dlopen(filename,RTLD_NOW)) == NULL) {
        aPtr = dlerror();
        printf("loadPlugin: dlopen failed, %s\n",aPtr);
        ptr->flags = PLUGIN_EMPTY;
        pluginFree(ptr);
        return -1;
     }

     if (xmpd_verbose)
        printf("loadPlugin: getting symbols dlopen\n");

     /* get symbols; plugins must implement these four */
     ptr->init = (int (*)())dlsym(ptr->solibhandle,"init");
     ptr->start = (int (*)())dlsym(ptr->solibhandle,"start");
     ptr->finish = (int (*)())dlsym(ptr->solibhandle,"finish");
     ptr->modinfo = (int (*)())dlsym(ptr->solibhandle,"modinfo");

     if ((ptr->init == NULL) || (ptr->start == NULL) ||
	 (ptr->finish == NULL) || (ptr->modinfo == NULL)) {
         printf("loadPlugin: plugin '%s' does not have required entry pts\n",
                 filename);
         ptr->flags = PLUGIN_EMPTY;
         dlclose(ptr->solibhandle);
         ptr->solibhandle = NULL;
         pluginFree(ptr);
         return -1;
     }

     strcpy(ptr->pathname,filename);
     ptr->loadTime = time(NULL);

     if (xmpd_verbose)
        printf("loadPlugin: calling modinfo\n");

     /* call modinfo for module name and description */
     (ptr->modinfo)(ptr->modulename,sizeof(ptr->modulename),
                    ptr->descr,sizeof(ptr->descr));

     if (xmpd_verbose)
        printf("loadPlugin: calling init\n");

     /* call init */
     if ((ret = (ptr->init)()) < 0) {
         printf("loadPlugin: plugin init failed, %d\n",ret);
         ptr->flags = PLUGIN_EMPTY;
         dlclose(ptr->solibhandle);
         ptr->solibhandle = NULL;
         pluginFree(ptr);
         return -1;
     }

     if (xmpd_verbose)
        printf("loadPlugin: calling start\n");

     /* call start */
     if ((ret = (ptr->start)()) < 0) {
        (ptr->finish)();
         printf("loadPlugin: plugin start failed, %d\n",ret);
         ptr->flags = PLUGIN_EMPTY;
         dlclose(ptr->solibhandle);
         ptr->solibhandle = NULL;
         pluginFree(ptr);
         return -1;
     }

     insertHead(pluginList,(void *)ptr);

     if (xmpd_verbose)
        printf("loadPlugin: finished\n");

     return 1;
}

#endif /* !WIN32 */

/* ************************************************************* */

#ifdef WIN32

static int loadPlugin(char *filename)
{
     plugin_t *ptr;
     int ret;
     struct _stat statbuf;

     if (filename == NULL) {
        return -1;
     }

     if (xmpd_verbose)
        printf("loadPlugin: checking %s\n",filename);

     /* check to see if plugin has already been added;
        if so, unload old one and load new one */
     if ((ptr = getPluginByFilename(filename)) != NULL) {

        if (ptr->flags & PLUGIN_UNLOADING) {
           if (xmpd_verbose)
              printf("loadPlugin: not loading %s (yet) since its unloading\n",
                     ptr->modulename);
	   return 0;
        }

        /* check to see if the file has been modified
           since we loaded it; if it has not, then
           dont bother re-loading it */

        if (_stat(filename,&statbuf) < 0) {
 	   perror("loadPlugin: stat of already loaded plugin failed");
           return -1;
        }

	/*    printf("loadPlugin: loaded plugin %s, times %ld %ld\n",
	      filename,statbuf.st_mtime, ptr->loadTime); */

        /* if has not been modified since loading, dont reload */
        if (statbuf.st_mtime < ptr->loadTime) {
           if (xmpd_verbose)
   	      printf("loadPlugin: %s has not been updated\n",filename);
 	   return 0;
        }

        if (ptr->flags & PLUGIN_VALID) {
           if (xmpd_verbose)
              printf("loadPlugin: going to unload %s\n",ptr->modulename);
           unloadPluginByPtr(ptr);
           return 0;
        }

        /* at this point, plugin is already unloaded, proceed */

     }
     else {
        ptr = find_unused_plugin();
     }

     /* now, ptr is unused/empty plugin struct */
     ptr->flags = PLUGIN_VALID;

     /* open shared library via dlopen */
     if ((ptr->solibhandle = LoadLibraryEx(filename,NULL,0)) == NULL) {
        printf("loadPlugin: LoadLibrary of %s, %d\n",filename,GetLastError());
        ptr->flags = PLUGIN_EMPTY;
        pluginFree(ptr);
        return -1;
     }

     /* get symbols; plugins must implement these four */
     ptr->init = GetProcAddress(ptr->solibhandle,"init");
     ptr->start = GetProcAddress(ptr->solibhandle,"start");
     ptr->finish = GetProcAddress(ptr->solibhandle,"finish");
     ptr->modinfo = GetProcAddress(ptr->solibhandle,"modinfo");

     if ((ptr->init == NULL) || (ptr->start == NULL) ||
	 (ptr->finish == NULL) || (ptr->modinfo == NULL)) {
         fprintf(stderr,
                 "loadPlugin: plugin '%s' does not have required entry pts\n",
                 filename);
         DBG4("loadPlugin: %lx %lx %lx %lx\n",
	      ptr->init, ptr->start, ptr->finish, ptr->modinfo);
         ptr->flags = PLUGIN_EMPTY;
         FreeLibrary(ptr->solibhandle);
         ptr->solibhandle = NULL;
         pluginFree(ptr);
         return -1;
     }

     strcpy(ptr->pathname,filename);
     ptr->loadTime = time(NULL);

     /* call modinfo for module name and description */
     (ptr->modinfo)(ptr->modulename,sizeof(ptr->modulename),
                    ptr->descr,sizeof(ptr->descr));

     /* call init */
     if ((ret = (ptr->init)()) < 0) {
         printf("loadPlugin: plugin init failed, %d\n",ret);
         ptr->flags = PLUGIN_EMPTY;
         FreeLibrary(ptr->solibhandle);
         ptr->solibhandle = NULL;
         pluginFree(ptr);
         return -1;
     }

     /* call start */
     if ((ret = (ptr->start)()) < 0) {
        (ptr->finish)();
         printf("loadPlugin: plugin start failed, %d\n",ret);
         ptr->flags = PLUGIN_EMPTY;
         FreeLibrary(ptr->solibhandle);
         ptr->solibhandle = NULL;
         pluginFree(ptr);
         return -1;
     }

     insertHead(pluginList,(void *)ptr);

     if (xmpd_verbose)
        printf("loadPlugin: finished\n");

     return 1;
}

#endif /* WIN32 */

/* ************************************************************* */
/* unload and un-initialize a plugin */

static int unloadPluginByName(char *modulename)
{
     plugin_t *p;
     int ret;
     linkedElement_t *e;

     if (modulename == NULL)
        return -1;

     pthread_mutex_lock(&pluginList->lock);
 
     e = pluginList->head;
     while (e != NULL) {
       p = (plugin_t *)e->data;
       if ((p) && (strcmp(modulename,p->modulename) == 0)) {
 	  ret = unloadPluginByPtr(p);
          pthread_mutex_unlock(&pluginList->lock);
          return ret;
       }

       e = e->next;
     }
     pthread_mutex_unlock(&pluginList->lock);

     return 0;
}

/* ************************************************************* */
/* called when engine is being shutdown; start the two-pass
   unload process */

int unloadAllPlugins()
{
     plugin_t *p;
     linkedElement_t *e;

     /* get lock */
     pthread_mutex_lock(&pluginList->lock);

     e = pluginList->head;
     while (e != NULL) {

       p = (plugin_t *)e->data;
       unloadPluginByPtr(p);

       e = e->next;
     }

     pthread_mutex_unlock(&pluginList->lock);

#ifndef WIN32
     sleep(30);
#else
     Sleep(30*1000);
#endif

     return 1;
}

/* ************************************************************* */
/* 
   called periodically to see if plugins need final
   cleanup 
*/

int cleanupAllPlugins()
{
     plugin_t *p;
     time_t now;
     linkedElement_t *e;

     /* get lock */
     now = time(NULL);
     pthread_mutex_lock(&pluginList->lock);

     e = pluginList->head;
     while (e != NULL) {

       p = (plugin_t *)e->data;
       if (p != NULL) {

          if ((p->flags & PLUGIN_UNLOADING) && 
              ((now - p->loadTime) > PLUGIN_CLEANUP_TIME) ) {
             unloadPluginByPtrFinal(p);
          }
       }

       e = e->next;
     }

     pthread_mutex_unlock(&pluginList->lock);

     return 1;
}

/* ************************************************************* */
/* find plugins in current-working dir and sub-dirs (arch-specific) */

#ifndef WIN32

int findLoadAllPlugins(char *dirname)
{
     DIR *currentdir;
     struct dirent *dent, *dptr;
     struct stat statbuf;
     char fullpathname[2048];
     u_char dbuf[sizeof(struct dirent)+FILENAME_MAX+1];

     if (dirname == NULL)
        return 0;

     DBG1("findLoadAllPlugins: starting in '%s'\n",dirname);

     if ((currentdir = opendir(dirname)) == NULL) {
        fprintf(stderr,"findLoadAllPlugins: could not open %s, %d\n",
                dirname,errno);
        return -1;
     }

     if (xmpd_verbose)
        printf("findLoadAllPlugins: starting\n");

     /* look for *.so files */
     dent = (struct dirent *)dbuf;
     while (readdir_r(currentdir,dent,&dptr) == 0) {
 
       /* sanity checks */
       if (dptr == NULL) {
	  break;
       }

       /* if entry name ends in .so, its a plugin; load it.
          strstr finds first occurrence, not last XXX */

       if (strstr(dent->d_name,".so") != NULL) {

	  /* skip things that say .so.old, or .so.new */
	  if (strstr(dent->d_name,".so.") != NULL)
	     continue;
	  if (strstr(dent->d_name,".so_") != NULL)
	     continue;
	  if (strstr(dent->d_name,"init.sol") != NULL)
	     continue;

	  /* DBG("findLoadAllPlugins: looking at '%s'\n",dent->d_name);*/

          sprintf(fullpathname,"%s/%s",dirname,dent->d_name);

	  /* if .new, do not load */ 
	  if (strstr(dent->d_name,".new") != NULL)
	     continue;
          /* if its lib*.so, dont load it since its a support library */
          if (strstr(dent->d_name,"lib") == dent->d_name)
	     continue;

          /* stat filename to see if its subdir or file */
	  if (stat(fullpathname,&statbuf) < 0) {
	     fprintf(stderr,"findLoadAllPlugins: stat %s failed, %d\n",
                     fullpathname,errno);
	     continue; /* with while loop */
	  }

          /* if its not a regular file */
          if (!S_ISREG(statbuf.st_mode)) {
	     continue; /* with while loop */
          }

          /* check if its for the right platform */

          if (strstr(dent->d_name,PLUGINSUFFIX))
             loadPlugin(fullpathname);

       } /* if .so */

       /* look for arch-specific sub-dirs?? -- XXX */

     } /* while */

     closedir(currentdir);

     if (xmpd_verbose)
        printf("findLoadAllPlugins: finished\n");
     return 1;
}

#endif /* !WIN32 */

#ifdef WIN32

int findLoadAllPlugins(char *dirname)
{
     HANDLE hFind;
     WIN32_FIND_DATA dent;
     struct _stat statbuf;
     char searchdir[1024];

     if (dirname == NULL)
        return 0;

     sprintf(searchdir,"%s\\*.*",dirname);

     DBG1("findLoadAllPlugins: starting for %s\n",dirname);

     hFind = FindFirstFile(searchdir,&dent);

     /* look for *.dll files */
     while (hFind != INVALID_HANDLE_VALUE) {

       DBG1("findLoadAllPlugins: examining %s\n",dent.cFileName);

       /* if entry name ends in .so, its a plugin; load it.
          strstr finds first occurrence, not last XXX */

       if (strstr(dent.cFileName,PLUGINSUFFIX) != NULL) {

	  /* if .new, do not load */ 
	  if (strstr(dent.cFileName,".new") != NULL) {
             if (FindNextFile(hFind,&dent) == 0) {
   	         break;
             }
	     continue;
 	  }
	  if (strstr(dent.cFileName,".old") != NULL) {
             if (FindNextFile(hFind,&dent) == 0) {
   	         break;
             }
	     continue;
 	  }
          /* if its a libblah.so, dont load it since its one of our
             support libs */
          if (strstr(dent.cFileName,"lib") == dent.cFileName) {
             if (FindNextFile(hFind,&dent) == 0) {
   	         break;
             }
	     continue;
          }    

 	  /* we could also look at the attributes in dent but
             we would have to rule out attributes
             rather than rule in attributes */

          /* stat filename to see if its subdir or file */
	  if (_stat(dent.cFileName,&statbuf) >= 0) {

             /* if its a regular file */
#define	_S_ISTYPE(mode, mask)	(((mode) & _S_IFMT) == (mask))
#define	S_ISREG(mode)	 _S_ISTYPE((mode), _S_IFREG)

             if (S_ISREG(statbuf.st_mode)) {
                loadPlugin(dent.cFileName);
             }

	  } /* if stat */

       } /* if -win32.dll */

       /* look for arch-specific sub-dirs?? -- XXX */

       if (FindNextFile(hFind,&dent) == 0) {
	  break;
       }

     } /* while */
 
     FindClose(hFind); 

     DBG("findLoadAllPlugins: finished\n");
     return 1;
}

#endif /* WIN32 */


/* ************************************************************* */

int getNumberPlugins()
{
     return numberElements(pluginList);

#if 0
     int ret = 0;

     pthread_mutex_lock(&plugin_lock);

     ptr = plugin_list;
     while (ptr != NULL) {
       if (ptr->flags != PLUGIN_EMPTY)
          ret++;

       ptr = ptr->next;
     }

     pthread_mutex_unlock(&plugin_lock);

     return ret;
#endif
}

/* ************************************************************* */
/* fetch plugin-table entries */

xmpErrorStatus_t var_plugList(char *key, xmpVariable_t *columns,
                              xmpPdu_t *out)
{
     char rowKey[256];
     xmpVariable_t *varptr, *cptr;
     u_int maxrows;
     int numRows = 0;
     int numVariables = 0;
     plugin_t *ptr;
     char timestr[32];
     linkedElement_t *e;

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     /* keys can be plugin-name, partial, *, \0 */
     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     pthread_mutex_lock(&pluginList->lock);

     /* cycle through plugin list */
     e = pluginList->head;
     while (e != NULL) {

       ptr = (plugin_t *)e->data;
       if (ptr == NULL) {
	  e = e->next;
          continue; /* with while loop */
       }

       /* row key is plugin name */
       sprintf(rowKey,"%s",ptr->modulename);

       if ((strcmp(rowKey,key) != 0) && (key[0] != 0) && 
           (key[0] != '*')) {
	  e = e->next;
 	  continue; /* with while plugins */
       }

       cptr = columns;
       while ((cptr != NULL) && (numRows < maxrows)) {

         /* malloc a return value */
         if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
             == NULL) {
	     perror("var_plugList: malloc failed");
             out->errorStatus = genError;
             /* release lock */
             pthread_mutex_unlock(&pluginList->lock);
             return genError;
         }
         bcopy(cptr,varptr,sizeof(xmpVariable_t));
         strcpy(varptr->varName.key,rowKey);

         switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {
	   case 1: /* name */
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            ptr->modulename);
              break;

   	   case 2: /* path */
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            ptr->pathname);
              break;

  	   case 3: /* descr */
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            ptr->descr);
              break;
  	   case 4: /* load time */
	      xmpConvertTimeStr(ptr->loadTime,timestr,sizeof(timestr));
              varptr->varSyntax = displayString;
              varptr->valueStrLen = sprintf(varptr->valueStr,"%s",timestr);
              break;

	   default:
	      fprintf(stderr,"var_plugList: invalid column\n");
              /* release lock */
              pthread_mutex_unlock(&pluginList->lock);
              out->errorStatus = noSuchName;
              out->errorIndex = numVariables;
              out->numberVariables = numVariables;
              return noSuchName;

         } /* switch column */

         /* glue into variable list */
         numVariables++;
         if (out->varListTail)
            out->varListTail->next = varptr;
         out->varListTail = varptr;
         if (out->varList == NULL)
            out->varList = varptr;           

         cptr = cptr->next;

       } /* while columns */

       numRows++;
       e = e->next;

     } /* while plugins */

     pthread_mutex_unlock(&pluginList->lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (xmpd_verbose)
        printf("var_plugList: %d variables, finished\n",numVariables);

     return noError;
}

/* ************************************************************* */
/* exported functions for plugins to access */

DLLExport int xmpGetAgentDebugLevel() { return xmpd_verbose; }

DLLExport int xmpGetAgentPort() { return xmpd_port; }

DLLExport float xmpGetAgentVersion() { return xmpd_version; }

/* ************************************************************* */

DLLExport int xmpConvertTimeStr(time_t time, char *timestr, int maxlen)
{
     char *ptr;

     if ((timestr == NULL) || (maxlen < 20))
        return -1;
     timestr[0] = 0;

     /* because we declare _POSIX_PTHREAD_SEMANTICS, ctime_r() now
        has same args as linux */
     ctime_r(&time,timestr);

#ifdef WIN32
     ptr = ctime(&time);
     strncpy(timestr,ptr,maxlen-1);
     timestr[maxlen-1] = 0;
#endif

     if (timestr[strlen(timestr)-1] == '\n')     
        timestr[strlen(timestr)-1] = 0; /* \n */

     return 1;
}

/* ************************************************************* */
/* plugins call this function to get a symbol value for 
   function pointers, variables, etc. from other plugins */

#ifndef WIN32 

DLLExport void *xmpGetSymbolValue(char *symbolname)
{
     static void *handle = NULL;
     void *retptr;
     plugin_t *ptr;
     linkedElement_t *e;

     if (handle == NULL) {
        handle = dlopen(NULL,RTLD_GLOBAL|RTLD_NOW);
        if (handle == NULL)
	   return NULL;
     }

     if (symbolname == NULL)
        return NULL;

     if ((retptr = dlsym(handle,symbolname)) != NULL)
        return retptr;

     /* otherwise, cycle through plugin list to find symbol */
     pthread_mutex_lock(&pluginList->lock);
     e = pluginList->head;
     retptr = NULL;
     while (e != NULL) {

       ptr = (plugin_t *)e->data;

       if ((ptr) && (ptr->flags == PLUGIN_VALID)) {
          retptr = dlsym(ptr->solibhandle,symbolname);
          if (retptr != NULL)
 	     break; /* from while loop */
       }
       e = e->next;
     }

     pthread_mutex_unlock(&pluginList->lock);

     return retptr;
}

#endif /* !WIN32 */

/* ************************************************************* */
/* plugins call this function to get a symbol value for 
   function pointers from other plugins */

#ifdef WIN32

DLLExport void *xmpGetSymbolValue(char *symbolname)
{
     static HMODULE handle = NULL;
     void *retptr = NULL;
     plugin_t *ptr;
     linkedElement_t *e;

     /* for this function, we may need to open each
        plugin/DLL and search for symbol, and FreeLibrary
        each opened library 
     */

     /* dont forget to set retptr to null!!! */

     pthread_mutex_lock(&pluginList->lock);
     e = pluginList->head;

     while (e != NULL) {
       ptr = (plugin_t *)e->data;
 
       if ((ptr) && (ptr->flags == PLUGIN_VALID)) {
 	  retptr = (void *)GetProcAddress(ptr->solibhandle,symbolname);
          if (retptr != NULL)
	     break;
       }
       e = e->next;
     }

     pthread_mutex_unlock(&pluginList->lock);
     return retptr;
}

#endif /* WIN32 */

/* ************************************************************* */
/* get our home directory */

DLLExport int xmpGetHomeDir(char *dirname, int maxlen)
{
     if ((dirname == NULL) || (maxlen < (strlen(xmpd_home_dir)+1)))
        return -1;

     strcpy(dirname,xmpd_home_dir);

     return 1;
}

/* ************************************************************* */

DLLExport int xmpGetSystemName(char *sysnamestr, int maxlen)
{
     if ((sysnamestr == NULL) || (maxlen < 10))
        return -1;

     /* initialize default based on valgrind feedback */
     sysnamestr[0] = 0;

     strncpy(sysnamestr,sysName,maxlen);
     sysnamestr[maxlen-1] = 0;

     return 1;
}

/* ************************************************************* */

DLLExport int xmpConvertIpAddr(u_int ipaddr, char *ipaddrstr, int maxlen)
{
     return convert_ipaddr(ipaddr,ipaddrstr,maxlen);
}

DLLExport int xmpConvertIpV6Addr(u_char *ipv6addr, char *ipv6str, int maxlen)
{
     /* ipv6addr is really a ptr to an in6_addr struct */
     return convert_ipv6addr(ipv6addr,ipv6str,maxlen);
}

/* ************************************************************* */
/* convert hostname or ascii IP addr to 32-bit IPv4 address */
/* use function from socklib.c */

DLLExport u_int xmpHostnameToIpaddr(char *host)
{
     return hostname_to_ipaddr(host);
}

/* ************************************************************* */

DLLExport int xmpdLog(char *msg)
{
     if (msg == NULL)
        return 0;

     return xmpd_log(msg);
}

/* ************************************************************* */

DLLExport int xmpAddressToHostname(char *addrstr, char *hostname, int maxlen)
{
     return addressToHostname(addrstr,hostname,maxlen);
}

DLLExport int xmpIpAddrToHostname(char *addrstr, char *hostname, int maxlen)
{
     return ipAddrToHostname(addrstr,hostname,maxlen);
}

/* ************************************************************* */

DLLExport int xmpAddModule(char *modulename, char *version, char *owner, 
                           char *dependList, int modtype)
{
     return addModule(modulename,version,owner,dependList,modtype);
}

/* ************************************************************* */

DLLExport int xmpRemoveModule(char *modulename)
{
     return removeModule(modulename);
}

/* ************************************************************* */

DLLExport int xmpCheckPeerModules(char *peerName)
{
     return checkPeerModules(peerName);
}

/* ************************************************************* */

DLLExport float xmpFindBinaryVersion(char *filename, char *specialStr)
{
     float findBinaryVersion();

     return findBinaryVersion(filename,specialStr);
}

/* ************************************************************* */

DLLExport xmpErrorStatus_t xmpSendTrap(char *ipaddr, xmpPdu_t *t, 
                                       char *eventType, char *sysObjID, 
                                       char *authenUser)
{
     return sendTrap(ipaddr,t,eventType,sysObjID,authenUser);
}


/* ************************************************************* */

#ifndef WIN32
DLLExport int xmpRegisterTrapHandler(char *name, int (*fptr)())
#else
DLLExport int xmpRegisterTrapHandler(char *name, FARPROC fptr)
#endif
{
     int registerTrapHandler();

     return registerTrapHandler(name,fptr);
}

/* ************************************************************* */
